/*
 * Copyright (C) 2025 RoyR. All rights reserved.
 * Distributed under the terms of the MIT License.
 *
 * Very Simple HTTP server
 * Version 1.0.1
 *   _   _   _   _   _   _
 *  / \ / \ / \ / \ / \ / \
 * ( v | s | H | T | T | D | (Very Simple HTTP Daemon)
 *  \_/ \_/ \_/ \_/ \_/ \_/
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <pwd.h>
#include <libgen.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#define BUFFER_SIZE 4096
#define VSHTTPD_VERSION "1.0.1"

static int client_fd;
static int server_fd;
static int log_enable = 0;
static int cgi_enable = 0;
static int vhost_enable = 0;
static char *main_docroot;
static short port = 8080;
static char *post_buffer = NULL;
static size_t post_buffer_length;
static size_t post_length = 0;
static char post_type[80+1];
static FILE *log;
extern int errno;

#ifndef IPV6
static struct sockaddr_in client_addr;
#else /* IPV6 is defined */
static struct sockaddr_in6 client_addr;
#endif

static char *getfile_type(char *filename) {
  int filename_length = strlen(filename);
  char *ext;
  while(filename_length > 0 && !(filename[filename_length] == '.')) {
    filename_length--;
  }
  ext = filename + filename_length;
  if(!strcmp(ext, ".html")) {
    return "text/html";
  } else if(!strcmp(ext, ".txt")) {
    return "text/plain";
  } else if(!strcmp(ext, ".jpg") || !strcmp(ext, ".jpe") || !strcmp(ext, ".jpeg")) {
    return "image/jpeg";
  } else if(!strcmp(ext, ".png")) {
    return "image/png";
  } else {
    return "binary/unknown";
  }
}

static void send_error(FILE *out, int id) {
  char *error_msg;
  char *http_msg = "HTTP/1.1";
  if(id == 404) { /* Not Found */
    fprintf(out, "%s 404 Not Found\r\n", http_msg);
    error_msg = "<h1>404 Not Found</h1>";
  } else if(id == 403) { /* Forbidden */
    fprintf(out, "%s 403 Forbidden\r\n", http_msg);
    error_msg = "<h1>403 Forbidden</h1>";
  } else if(id == 400) {
    fprintf(out, "%s 400 Bad Request\r\n", http_msg);
    error_msg = "<h1>400 Bad Request</h1>";
  } else if(id == 501) {
    fprintf(out, "%s 501 Method Not Implemented\r\n", http_msg);
    error_msg = "<h1>Method Not Implemented</h1>";
  } else {
    fprintf(out, "%s 500 Internal Server Error\r\n", http_msg);
    error_msg = "<h1>500 Internal Server Error</h1>\r\n";
  }
  fprintf(out, "Server: vshttpd\r\n");
  fprintf(out, "Content-Type: text/html\r\n");
  fprintf(out, "\r\n");
  fprintf(out, "%s", error_msg);
}

static void send_200(FILE *out, char *content_type, off_t size) {
  fprintf(out, "HTTP/1.1 200 OK\r\n");
  fprintf(out, "Server: vshttpd\r\n");
  fprintf(out, "Connection: close\r\n");
  fprintf(out, "Content-Length: %zu\r\n", (size_t)size);
  fprintf(out, "Content-Type: %s\r\n\r\n", content_type);
}

static char *build_env(char const *name, char const *value) {
  char *ptr = malloc((strlen(name) + strlen(value)) + 2);

  if(ptr) {
    strcpy(ptr, name);
    strcat(ptr, "=");
    strcat(ptr, value);
  } else {
    fprintf(log, "Failed to allocate memory for environment variable: %s\n",
	    strerror(errno));
    exit(1);
  }
  return ptr;
}

static void format_ip(void *sockaddr_ptr, char *ip_addr) {
  const char *buf_ptr;
  char buf[50];
  
#ifndef IPV6
  struct in_addr *addr;
  struct sockaddr_in *sockaddr;
  int af = AF_INET;
  
  sockaddr = (struct sockaddr_in *)sockaddr_ptr;
  addr = &sockaddr->sin_addr;
#else
  struct in6_addr *addr;
  struct sockaddr_in6 *sockaddr;
  int af = AF_INET6;
  
  sockaddr = (struct sockaddr_in6 *)sockaddr_ptr;
  addr = &sockaddr->sin6_addr;
#endif
  buf_ptr = inet_ntop(af, (void *)addr, buf, sizeof(buf));
  
  if(buf_ptr) {
    strcpy(ip_addr, buf_ptr);
  }
  return;
}

/* The 'strdup' function is not available in the C 11 or later standards  */
static char *our_strdup(const char *s)
{
  size_t size = strlen(s) + 1;
  char *p = malloc(size);
  if(p) {
    memcpy(p, s, size);
  }
  return p;
}

static void execute_cgi(FILE *out, char *uri, const char *method, const char *doc_root, const char *host) {
  char *argv[256+1] = { NULL };
  char *envp[15+1] = { NULL };
  int env_num = 0;
  FILE *cgi_out;
  FILE *cgi_in;
  char buffer[BUFSIZ];
  int cgi_out_pipe[2];
  int cgi_in_pipe[2];
  int bytes_read, bytes_written;
  char exec_path[4096];
  char *cgi_bin;
  char *cgi_bin_ptr;
  char ip_addr[50];
  char port_str[5+1];
  int argv_index = 1;

  if(uri[strlen(uri) - 1] == '/') {
    send_error(out, 403);
    return;
  }
  if((pipe(cgi_out_pipe) < 0) || (pipe(cgi_in_pipe) < 0)) {
    send_error(out, 500);
    fprintf(log, "pipe: %s\n", strerror(errno));
    return;
  }

  sprintf(port_str, "%hu", port);
  
  buffer[0] = '\0';
  strncat(buffer, uri, strcspn(uri, "?"));

  format_ip(&client_addr, ip_addr);
  envp[env_num] = build_env("DOCUMENT_ROOT", doc_root); env_num++;
  envp[env_num] = build_env("GATEWAY_INTERFACE", "CGI/1.1"); env_num++;
  envp[env_num] = build_env("REQUEST_URI", uri); env_num++;
  envp[env_num] = build_env("SERVER_SOFTWARE", "vshttpd (Unix)"); env_num++;
  envp[env_num] = build_env("HTTP_HOST", host); env_num++;
  envp[env_num] = build_env("QUERY_STRING", uri + strcspn(uri, "?") + 1); env_num++;
  envp[env_num] = build_env("REQUEST_METHOD", method); env_num++;
  envp[env_num] = build_env("REMOTE_ADDR", ip_addr); env_num++;
  envp[env_num] = build_env("SCRIPT_NAME", buffer); env_num++;
  envp[env_num] = build_env("SERVER_NAME", host); env_num++;
  envp[env_num] = build_env("SERVER_PORT", port_str); env_num++;
  envp[env_num] = build_env("SERVER_PROTOCOL", "HTTP/1.1"); env_num++;
  if(!strcmp(method, "POST")) {
    char content_length[25];
    sprintf(content_length, "%zu", post_length);
    envp[env_num] = build_env("CONTENT_LENGTH", content_length); env_num++;
    envp[env_num] = build_env("CONTENT_TYPE", post_type); env_num++;
  }
  envp[env_num] = NULL;

  if(strchr(uri, '?')) {
    strcpy(exec_path, doc_root);
    strncat(exec_path, uri, strcspn(uri, "?"));
    int counter = 0 + strcspn(uri, "?");
    while(uri[counter] != 0 && counter < 256) {
      if(uri[counter] == '&' || uri[counter] == '?') {
    	argv[argv_index] = &uri[counter+1];
	uri[counter] = 0;
    	argv_index++;
      }
      counter++;
    }
  } else {
    strcpy(exec_path, doc_root);
    strcat(exec_path, uri);
  }
  cgi_bin_ptr = our_strdup(exec_path);
  if(!cgi_bin_ptr) {
    fprintf(log, "malloc: %s\n", strerror(errno));
    send_error(out, 500);
    return;
  }
  cgi_bin = dirname(cgi_bin_ptr);
  argv[0] = exec_path;
  argv[argv_index] = NULL;
  if(log_enable)
    fprintf(log, "Running CGI: %s, for host: %s\n", exec_path, host);
  pid_t pid = fork();
  if(pid == 0) {
    /* Child process */
    fclose(out);
    dup2(cgi_out_pipe[1], STDOUT_FILENO);/* Redirect stdout to the pipe */
    dup2(cgi_in_pipe[0], STDIN_FILENO); /* Redirect stdin from the in pipe */
    dup2(fileno(log), STDERR_FILENO); /* Redirect stderr to log */
    close(cgi_in_pipe[1]);
    close(cgi_out_pipe[0]);
    if(chdir(cgi_bin) < 0) {
      fprintf(stdout, "Content-Type: text/html\r\n\r\n");
      fprintf(stdout, "<h1>500 Internal Server Error</h1>");
      fprintf(log, "chdir(%s): %s\n", cgi_bin, strerror(errno));
      exit(1);
    }
    execve(exec_path, argv, envp); /* Execute the CGI script */
    fprintf(log, "%s: %s\n", exec_path, strerror(errno));
    fprintf(stdout, "Content-Type: text/html\r\n\r\n");
    fprintf(stdout, "<h1>500 Internal Server Error</h1>");
    exit(1); /* Only reached if there is a problem with th execve */
  } else if(pid > 0) {
    /* Parent process */
    env_num = 0;
    while(envp[env_num]) {
      free(envp[env_num]);
      env_num++;
    }
    free(cgi_bin_ptr);
    close(cgi_in_pipe[0]); /* Close the unneed ends of the pipes */
    close(cgi_out_pipe[1]);
    cgi_out = fdopen(cgi_out_pipe[0], "r");
    cgi_in = fdopen(cgi_in_pipe[1], "w");
    if(!cgi_in || !cgi_out) {
      send_error(out, 500);
      fprintf(log, "fdopen: %s\n", strerror(errno));
      return;
    }

    if(!strcmp(method, "POST")) {
      bytes_read = 0;
      if(post_buffer_length > fwrite(post_buffer, 1, \
				     post_buffer_length, cgi_in)) {
	send_error(out, 500);
	fclose(cgi_out);
	fclose(cgi_in);
	return;
      }
      size_t byte_counter = 0;
      if(post_buffer_length < post_length) {
	bytes_written = 1;
	while(byte_counter < (post_length - post_buffer_length)) {
	  bytes_read = recv(client_fd, buffer, sizeof(buffer), 0);
	  if(bytes_read > 0 && bytes_written > 0) {
	    bytes_written = fwrite(buffer, 1, bytes_read, cgi_in);
	    byte_counter += bytes_read;
	  } else {
	    break;
	  }
	}
      }
    }
    fclose(cgi_in);
    fprintf(out, "HTTP/1.1 200 OK\r\n");
    fprintf(out, "Server: vshttpd\r\n");
    while(fgets(buffer, BUFSIZ, cgi_out) != NULL) {
      fputs(buffer, out);
      if((strstr(buffer, "\r\n") == buffer) ||
	 (strstr(buffer, "\n") == buffer))
	break;
    }
    if(strcmp(method, "HEAD")) {
      bytes_read = fread(&buffer, 1, sizeof(buffer), cgi_out);
      bytes_written = 1;
      while (bytes_read > 0 && bytes_written > 0) {
	bytes_written = fwrite(&buffer, 1, bytes_read, out);
	bytes_read = fread(&buffer, 1, sizeof(buffer), cgi_out);
      }
    }
    fclose(cgi_out);
    wait(NULL); /* Wait for the child to exit */
  } else {
    /* fork() failed */
    fprintf(log, "fork: %s\n", strerror(errno));
    send_error(out, 500);
    env_num = 0;
    while(envp[env_num]) {
      free(envp[env_num]);
      env_num++;
    }
  }
}

static void serve_file(FILE *out, const char *filename, const char *method) {
  FILE *file = fopen(filename, "rb");
  struct stat file_stats;
  size_t bytes_read;
  size_t bytes_writen;

  if(strstr(filename, "/cgi-bin/")) {
    send_error(out, 403);
    return;
  }
  if(!file || stat(filename, &file_stats)) {
    if(errno == ENOENT) {
      send_error(out, 404);
    } else {
      send_error(out, 403);
    }
    return;
  }

  send_200(out, getfile_type((char *)filename), file_stats.st_size);
  if(strcmp(method, "HEAD")) {
    char buffer[BUFFER_SIZE];
    bytes_read = fread(&buffer, 1, sizeof(buffer), file);
    bytes_writen = 1;
    while (bytes_read > 0 && bytes_writen > 0) {
      bytes_writen = fwrite(buffer, 1, bytes_read, out);
      bytes_read = fread(buffer, 1, sizeof(buffer), file);
    }
  }
  fclose(file);
}

static void send_redirect(FILE *out, int type, const char *url) {
  fprintf(out, "HTTP/1.1 %i Moved Permanently\r\n", type);
  fprintf(out, "Server: vshttpd\r\n");
  fprintf(out, "Location: %s\r\n", url);
  fprintf(out, "Content-Type: text/html\r\n\r\n");
  if(type == 301)
    fprintf(out, "<h1>301 Moved Permanently</h1>\r\n");
  else
    fprintf(out, "<h1>%i Redirect</h1>\r\n", type);
  return;
}

static void handle_request(FILE *out, const char *host, const char *request) {
  char method[8+1];
  char uri[2048+1];
  char path[4096+1];
  char docroot[4096+1];
  char ip_addr[50+1];
  struct stat file_stats;

  if(log_enable)
    format_ip(&client_addr, ip_addr);
  
  sscanf(request, "%8s %2048s", method, uri);
  if(strstr(host, "/..") || strstr(uri, "/..")) {
    send_error(out, 400);
    return;
  }
  if(uri[0] != '/') {
    send_error(out, 400);
    return;
  }

  strcpy(path, main_docroot);
  if(vhost_enable) {
    strcat(path, "/");
    strcat(path, host);
  }

  strcpy(docroot, path);
  strcat(path, uri);

  if(stat(path, &file_stats) && strstr(uri, "/cgi-bin/") != uri) {
    if(errno == ENOENT) {
      send_error(out, 404);
    } else {
      send_error(out, 403);
    }
    return;
  }

  /* Simple virtual hosting based on the hostname */
  if(!access(docroot, F_OK)) {
    if(access(docroot, R_OK) < 0) {
      /* 403 Forbidden */
      send_error(out, 403);
      return;
    }
    if(log_enable) {
      fprintf(log, "%s, %s %s for host: %s\n", ip_addr, method, \
	      uri, host);
    }
    if(strcmp(method, "GET") && strcmp(method, "HEAD") && \
       strcmp(method, "POST")) {
      send_error(out, 501);
      return;
    }
    if(strstr(uri, "/cgi-bin/") == uri) {
      if(cgi_enable)
	execute_cgi(out, uri, method, docroot, host);
      else
	send_error(out, 403);
    } else if(strcmp(method, "POST")) {
      if(path[strlen(path)-1] == '/') {
	strcat(path, "index.html");
      } else if(S_ISDIR(file_stats.st_mode)) {
	if(strlen(host)) {
	  strcpy(path, "http://");
	  strcat(path, host);
	} else {
	  path[0] = '\0';
	}
	strcat(path, uri);
	strcat(path, "/");
	send_redirect(out, 301, path);
	return;
      }
      serve_file(out, path, method);
    } else {
      send_error(out, 400);
    }
  } else {
    send_error(out, 404);
  }
}

static void handle_client(int client_fd) {
  char buffer[BUFFER_SIZE];
  FILE *out = fdopen(client_fd, "w");

  ssize_t buffer_index = 0;
  ssize_t bytes_read = 0;

  memset(buffer, 0, sizeof(buffer));
  bytes_read = recv(client_fd, buffer, sizeof(buffer) - 1, 0);
  buffer_index = bytes_read;
  if(buffer_index > 0) {
    while(!strstr(buffer, "\r\n\r\n") && !strstr(buffer, "\n\n") && \
	  (sizeof(buffer) - buffer_index) - 1 > 1) {
      bytes_read = recv(client_fd, buffer + buffer_index, sizeof(buffer) - \
			buffer_index - 1, 0);
      if(bytes_read > 0) {
	buffer_index += bytes_read;
      } else {
	break;
      }
    }
    char host[256+1];
    host[0] = '\0';
    char *host_header = strstr(buffer, "Host: ");
    if(host_header) {
      sscanf(host_header + 6, "%256s", host);
    } else if(!vhost_enable) {
      host[0] = '\0';
    } else {
      send_error(out, 400);
      fclose(out);
      return;
    }
    if(strstr(buffer, "POST ") == buffer) {
      char *content_length_header = strstr(buffer, "Content-Length: ");
      if(!content_length_header)
	content_length_header = strstr(buffer, "Content-length: ");
      char *content_type_header = strstr(buffer, "Content-Type: ");
      if(!content_type_header)
	content_type_header = strstr(buffer, "Content-type: ");
      if(content_type_header)
	sscanf(content_type_header + 14, "%80s", post_type);
      else
	strcpy(post_type, "binary/unkown");

      size_t content_length = 0;
      if(content_length_header) {
	sscanf(content_length_header + 16, "%zu", \
	       &content_length);
      } else {
	send_error(out, 400);
	fclose(out);
	return;
      }
      char *headers_end;
      headers_end = strstr(buffer, "\r\n\r\n");
      if(!headers_end) {
	headers_end = strstr(buffer, "\n\n"); headers_end += 2;
      } else {
	headers_end += 4;
      }
      if(headers_end && (headers_end < buffer + buffer_index)) {
	post_buffer = headers_end;
	post_buffer_length = (buffer + buffer_index) - headers_end;
      }
      post_length = content_length;
    }
    handle_request(out, host, buffer);
  }
  fclose(out);
  close(client_fd);
}

static void sigpipe_handler(int sig) {
  close(client_fd);
  exit(sig);
}

static void usage(char *name) {
  fprintf(stderr, "Usage: %s [-hlpv] [-d docroot] [-p port] [-u user]\n", name);
  fprintf(stderr, "-b\t\tFork off into the background and become a daemon\n");
  fprintf(stderr, "-c\t\tEnable CGI\n");
  fprintf(stderr, "-d docroot\tSet the document root (default /var/www/html)\n");
  fprintf(stderr, "-h\t\tShow this help\n");
  fprintf(stderr, "-L logfile\tEnable logging to logfile (only for use with with -b)\n");
  fprintf(stderr, "-l\t\tEnable logging to STDERR (only for use without -b)\n");
  fprintf(stderr, "-p port\t\tPort number to listen on (default 8080)\n");
  fprintf(stderr, "-u user\t\tUser to switch to if run by root (default httpd)\n");
  fprintf(stderr, "-V\t\tEnable virtual hosting\n");
  fprintf(stderr, "-v\t\tShow copyright and version infomation\n");
  return;
}

static void version(void) {
  fprintf(stderr, "vshttpd " VSHTTPD_VERSION "\n");
  fprintf(stderr, "Copyright (C) 2024, 2025 RoyR. All rights reserved.\n");
  fprintf(stderr, "Distributed under the terms of the MIT License.\n");
  return;
}

static void shutdown_handler(int sig) {
  if((sig == SIGINT) || (sig == SIGTERM)) {
    if(log_enable)
      fprintf(log, "Server shutdown requested.\n");
    fclose(log);
    close(server_fd);
    exit(sig);
  }
}

int main(int argc, char *argv[]) {
  int optval = 1;
  int background = 0;
  char *logfile = NULL;
  int logfile_enable = 0;
  int log_fd;
#ifndef IPV6
  struct sockaddr_in server_addr;
  short af = AF_INET;
#else /* IPV6 is defined */
  struct sockaddr_in6 server_addr;
  short af = AF_INET6;
#endif
  int opt;
  char *user = "httpd";
  main_docroot = "/var/www/html";
  extern char *optarg;

  while((opt = getopt(argc, argv, "bchd:L:lp:u:Vv")) != -1) {
    switch(opt) {
    case 'b':
      background = 1;
      break;
    case 'c':
      cgi_enable = 1;
      break;
    case 'V':
      vhost_enable = 1;
      break;
    case 'L':
      logfile = optarg;
      logfile_enable = 1;
      break;
    case 'l':
      log_enable = 1;
      break;
    case 'p':
      if(!(port = atoi(optarg))) {
	fprintf(stderr, "port is an integer in the range 1-65536\n");
	return 1;
      }
      break;
    case 'd':
      main_docroot = optarg;
      break;
    case 'u':
      user = optarg;
      break;
    case 'h':
      usage(argv[0]);
      return 1;
    case 'v':
      version();
      return 1;
    default:
      usage(argv[0]);
      return 1;
    }
  }
  if(optind < argc) {
    usage(argv[0]);
    return 1;
  }
  if((logfile_enable && !background) || (log_enable && background)) {
    usage(argv[0]);
    return 1;
  }
  if(logfile_enable) {
    mode_t mode = S_IRUSR|S_IWUSR|S_IRGRP;
    log_fd = open(logfile, O_APPEND|O_CREAT|O_WRONLY, mode);
    if(log_fd < 0) {
      perror(logfile);
      return 1;
    }
    close(log_fd);
  }

  close(STDIN_FILENO);
  int nullfd = open("/dev/null", O_RDWR);
  if(nullfd != 0) {
    fprintf(stderr, "Could not re-open STDIN to /dev/null\n");
    return 0;
  }
    
  if(background) {
    switch(fork())
      {
      case -1:
	perror("fork");
	return 1;
      case 0:
	break;
      default:
	exit(EXIT_SUCCESS); /* Exit parent process */
      }
    if(setsid() == -1) { /* Set this process to leader of new session */
      perror("setsid");
      return 1;
    }
    switch(fork())
      {
      case -1:
	perror("fork");
	return 1;
      case 0:
	break;
      default:
	exit(EXIT_SUCCESS); /* Exit parent process */
      }
    umask(S_IWGRP|S_IWOTH);
    int max_fd = sysconf(_SC_OPEN_MAX);
    if(max_fd == -1)
      max_fd = 8192;
    if(logfile_enable) {
      if(access(logfile, W_OK) < 0) {
	perror(logfile);
	return 1;
      }
    }
    while(max_fd > 0) {
      close(max_fd);
      max_fd--;
    }
    close(STDIN_FILENO);
    int fd = open("/dev/null", O_RDWR);
    if(fd != STDIN_FILENO)
      return 1;
    if(dup2(STDIN_FILENO, STDOUT_FILENO) != STDOUT_FILENO)
      return 1;
    if(dup2(STDIN_FILENO, STDERR_FILENO) != STDERR_FILENO)
      return 1;
    if(logfile_enable) {
      log_fd = open(logfile, O_APPEND|O_WRONLY);
      if(log_fd < 0) {
	return 1;
      }
      log = fdopen(log_fd, "ab");
    }
    chdir("/");
    log_enable = 1;
  } else {
    log = fdopen(fileno(stderr), "w");
  }
  if(setvbuf(log, NULL, _IONBF, 0) != 0) {
    fprintf(log, "setvbuf: %s\n", strerror(errno));
  }
  socklen_t client_len = sizeof(client_addr);
  struct passwd *passwd_info;

  /* Create socket */
  if ((server_fd = socket(af, SOCK_STREAM, 0)) < 0) {
    fprintf(log, "Failed to create socket: %s\n", strerror(errno));
    return 1;
  }
  
  /* Set up server address and bind */
  memset(&server_addr, 0, sizeof(server_addr));
#ifndef IPV6
  server_addr.sin_family = af;
  server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  server_addr.sin_port = htons(port);
#else
  server_addr.sin6_family = af;
  server_addr.sin6_addr = in6addr_any;
  server_addr.sin6_port = htons(port);
#endif
  
  if(setsockopt(server_fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) != 0) {
    fprintf(log, "Failed to set socket options: %s\n", strerror(errno));
    return 1;
  }
  
  if(bind(server_fd, (struct sockaddr *)&server_addr, sizeof(server_addr)) < 0) {
    fprintf(log, "Failed to bind socket: %s\n", strerror(errno));
    return 1;
  }
  
  if(listen(server_fd, 10) > 0) {
    fprintf(log, "Failed to listen on socket!: %s\n", strerror(errno));
    return 1;
  }
  
  if(getuid() == 0) {
    passwd_info = getpwnam(user);
    if(passwd_info == NULL) {
      fprintf(log, "User \"%s\" does not exist\n", user);
      return 1;
    }
    setuid(passwd_info->pw_uid);
    setgid(passwd_info->pw_gid);
  }

  signal(SIGPIPE, sigpipe_handler);
  signal(SIGCHLD, SIG_IGN);
  signal(SIGINT, shutdown_handler);
  signal(SIGTERM, shutdown_handler);

  fprintf(log, "Server listening on port %d\n", port);

  while(1) {
    client_fd = accept(server_fd, (struct sockaddr *)&client_addr, \
		       &client_len);
    if (client_fd < 0) {
      fprintf(log, "Failed to accept connection!: %s\n", strerror(errno));
      continue;
    }
    if(fork() == 0) {
      close(server_fd);
      signal(SIGCHLD, SIG_DFL);
      handle_client(client_fd);
      fclose(log);
      exit(0);
    } else {
      close(client_fd);
    }
  }
  return 0;
}
